/*
** ADM-XRC Series Driver
**
** (C) Copright 2006 Alpha Data
**
** methods_lbrst.c - Model-specific routines for performing a local bus
**                   reset.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

/*
** -----------------------------------------------------------------------
** MethodLocalBusReset
** -----------------------------------------------------------------------
*/

MethodResult
methodLocalBusReset9080(
  DeviceContext* pDevCtx,
  os_boolean bAssert)
{
  return MethodResult_NotSupported;
}

MethodResult
methodLocalBusReset9656(
  DeviceContext* pDevCtx,
  os_boolean bAssert)
{
  OS_SPINLOCK_FLAGS f;
  PLX9656REGS* pRegs;

  pRegs = pDevCtx->hw.adapter.pci9656;
  os_acquire_spinlock(&pDevCtx->shadow.lock, &f);
  if (bAssert) {
    pDevCtx->shadow.regs.plxbased.cntrl |= (0x1U << 30);
  } else {
    pDevCtx->shadow.regs.plxbased.cntrl &= ~(0x1U << 30);
  }
  os_write_register_le32(&pRegs->CNTRL, pDevCtx->shadow.regs.plxbased.cntrl);
  os_release_spinlock(&pDevCtx->shadow.lock, f);

  return MethodResult_Success;
}

MethodResult
methodLocalBusResetXpl(
  DeviceContext* pDevCtx,
  os_boolean bAssert)
{
  OS_SPINLOCK_FLAGS f;
  XPLBRIDGEREGS* pRegs;
  os_uint32 lctl;

  pRegs = pDevCtx->hw.adapter.xpl;
  os_acquire_spinlock(&pDevCtx->shadow.lock, &f);
  lctl = os_read_register_le32(&pRegs->LCTL);
  if (bAssert) {
    lctl |= (0x1U << 0);
  } else {
    lctl &= ~(0x1U << 0);
  }
  os_write_register_le32(&pRegs->LCTL, lctl);
  os_release_spinlock(&pDevCtx->shadow.lock, f);

  return MethodResult_Success;
}
