static void
initCard9080(
  DeviceContext* pDevCtx)
{
  os_dbgprint(1, ("initCard9080: entered\n"));

  /* Enable Direct Master access */
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMRR,   0xFF800000U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMLBAM, 0xFF000000U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMLBAI, 0x0U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMPBAM, 0x1U);

  if (admxrcDriverParams.pciReadMultiple) {
    /* Use PCI read multiple commands when a bus master */
    os_dbgprint(0, ("initCard9080: selecting PCI read multiple command for bus master reads\n"));
    pDevCtx->shadow.regs.plxbased.cntrl = os_read_register_le32(&pDevCtx->hw.adapter.pci9080->CNTRL);
    pDevCtx->shadow.regs.plxbased.cntrl &= ~0xf0fU;
    pDevCtx->shadow.regs.plxbased.cntrl |= 0xc0cU;
    os_write_register_le32(&pDevCtx->hw.adapter.pci9080->CNTRL, pDevCtx->shadow.regs.plxbased.cntrl);
  }
}

static void
initCard9656(
  DeviceContext* pDevCtx)
{
  /*
  ** Currently, as far as this driver is concerned, card initialization
  ** for the PCI9656 is the same as for the PCI9080, but with a few extra
  ** bits and pieces.
  */
  initCard9080(pDevCtx);

  if (ADMXRC_9656_ERRATA_WORKAROUND) {
    os_uint32 val;

    os_dbgprint(1, ("initCard9656: using PCI9656 rev AC/AD workarounds\n"));

    /*
    ** Implement workarounds for PCI9656 rev AC and AD errata
    */
    
    /* DMPBAM register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->DMPBAM);
    val &= ~(0x1U <<  2); /* Disable Direct Master read-ahead */
    /* Do not prefetch more than 8 QWORDs for a Direct Master read */
    val &= ~(0x1U <<  3);
    val |=  (0x1U << 12);
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->DMPBAM, val);
    
    /* EROMBA register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->EROMBA);
    val &= ~(0x1U <<  4); /* Don't use local bus backoff for Direct Master transfers */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->EROMBA, val);
    
    /* LBRD0 register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD0);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 0 reads */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD0, val);
    
    /* LBRD1 register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD1);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 1 reads */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD1, val);
    
    /* LMISC2 register */
    val = os_read_register_le8(&pDevCtx->hw.adapter.pci9656->LMISC2);
    val |=  (0x1U <<  5); /* Reserve 1 location in Direct Slave write FIFO */
    os_write_register_le8(&pDevCtx->hw.adapter.pci9656->LMISC2, val);
    
    /* MARBR register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->MARBR);
    val &= ~(0x1U << 16); /* Disable local bus latency timer */
    val &= ~(0x1U << 23); /* Assert REQ# on PCI for entire DMA transfer */
    val |=  (0x1U << 24); /* Operate in PCI 2.1 mode */
    val |=  (0x1U << 26); /* Flush Direct Slave read FIFO after Direct Slave write */
    val &= ~(0x1U << 28); /* Direct slave read ahead mode disabled */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->MARBR, val);
  }
}

static void
initCard(
  DeviceContext* pDevCtx)
{
  os_dbgprint(1, ("initCard: entered\n"));
  
  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    initCard9080(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LS:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC4SX:
    initCard9656(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXPL_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXPL_PCI:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCI:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCIX:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCI:
  case PCI_SUBDEVICE_ADATA_ADMXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADCPXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADPEXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5T1:
  case PCI_SUBDEVICE_ADATA_ADMXRC5T2:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LXA:
  case PCI_SUBDEVICE_ADATA_ADMAMC5A2:
    /* Nothing to do yet */
    break;

  default:
    os_dbgprint(0, ("*** initCard: unrecognised hardware ID: vId=0x%4.4lx dId=0x%4.4lx subVid=0x%4.4lx subDevice0x%4.4lx revision=0x%2.2lx\n",
		    (unsigned long) pDevCtx->devId.pci.vendor,
		    (unsigned long) pDevCtx->devId.pci.device,
		    (unsigned long) pDevCtx->devId.pci.subVendor,
		    (unsigned long) pDevCtx->devId.pci.subDevice,
		    (unsigned long) pDevCtx->devId.pci.revision));
    break;
  }
}
